@@ -1,4 +1,4 @@
-use crate::format::{self, Format};
+use crate::format::{Format, Formatter};
 use crate::items::Form;
 use crate::parse::{self, Parse, TokenStream};
 use crate::span::{Position, Span};
@@ -13,22 +13,21 @@ pub struct Module {
 
 impl Parse for Module {
     fn parse(ts: &mut TokenStream) -> parse::Result<Self> {
-        let sof = ts.current_whitespace_token()?.start_position();
+        let sof = ts.prev_token_end_position();
         let mut forms = Vec::new();
         while !ts.is_eof()? {
             forms.push(ts.parse()?);
         }
-        let eof = ts.current_whitespace_token()?.end_position();
+        let eof = ts.next_token_start_position()?;
         Ok(Self { sof, forms, eof })
     }
 }
 
 impl Format for Module {
-    fn format(&self, fmt: &mut format::Formatter) -> format::Result<()> {
+    fn format(&self, fmt: &mut Formatter) {
         for form in &self.forms {
-            form.format(fmt)?;
-            fmt.write_newline()?;
+            form.format(fmt);
+            fmt.add_newline();
         }
-        Ok(())
     }
 }
