@@ -1,63 +1,41 @@
+use crate::items::components::Either;
 use crate::items::forms::{DefineDirective, IncludeDirective};
-use crate::items::generics::Either;
 use crate::items::macros::{Macro, MacroName};
 use crate::items::symbols::{OpenParenSymbol, QuestionSymbol};
 use crate::items::tokens::{
     AtomToken, CharToken, CommentKind, CommentToken, FloatToken, IntegerToken, KeywordToken,
-    StringToken, SymbolToken, Token, VariableToken, WhitespaceToken,
+    LexicalToken, StringToken, SymbolToken, VariableToken,
 };
-use crate::items::Module;
-use crate::parse::{Parse, Result, ResumeParse};
+use crate::parse::include::IncludeHandler;
+use crate::parse::{Error, IncludeOptions, Parse, Result, ResumeParse};
 use crate::span::{Position, Span as _};
 use erl_tokenize::values::Symbol;
 use erl_tokenize::{PositionRange as _, Tokenizer};
 use std::collections::{BTreeMap, HashSet};
-use std::io::{BufReader, BufWriter};
 use std::path::PathBuf;
 use std::sync::Arc;
 
-#[derive(Debug, Default, Clone)]
-pub struct TokenStreamOptions {
-    include_dirs: Vec<PathBuf>,
-    include_cache_dir: Option<PathBuf>,
-}
-
-impl TokenStreamOptions {
-    pub fn new() -> Self {
-        Self::default()
-    }
-
-    pub fn include_dirs(mut self, dirs: Vec<PathBuf>) -> Self {
-        self.include_dirs = dirs;
-        self
-    }
-
-    pub fn include_cache_dir(mut self, dir: PathBuf) -> Self {
-        self.include_cache_dir = Some(dir);
-        self
-    }
-}
-
 #[derive(Debug)]
 pub struct TokenStream {
     tokenizer: Tokenizer<String>,
-    tokens: Vec<Token>,
+    tokens: Vec<LexicalToken>,
     current_token_index: usize,
     comments: BTreeMap<Position, CommentToken>,
     macros: BTreeMap<Position, Macro>,
-    macro_defines: BTreeMap<MacroDefineKey, MacroDefine>,
+    macro_defines: MacroDefines,
+    new_macro_defines: HashSet<MacroDefineKey>,
     missing_macros: HashSet<String>,
-    known_replacement: HashSet<(usize, Vec<Token>)>,
-    included: HashSet<String>,
-    parsing_macro_replacement: bool,
-    parsing_macro_arg: bool,
-    options: TokenStreamOptions,
+    known_replacement: HashSet<(usize, Vec<LexicalToken>)>,
+    disable_macro_expand: bool,
+    parsing_tokens: bool,
     text: Arc<String>,
     path: Option<Arc<PathBuf>>,
+    last_parse_error: Option<Error>,
+    include: IncludeHandler,
 }
 
 impl TokenStream {
-    pub fn new(tokenizer: Tokenizer<String>, options: TokenStreamOptions) -> Self {
+    pub fn new(tokenizer: Tokenizer<String>, options: IncludeOptions) -> Self {
         let text = Arc::new(tokenizer.text().to_owned());
         let path = tokenizer
             .next_position()
@@ -70,26 +48,82 @@ impl TokenStream {
             comments: BTreeMap::new(),
             macros: BTreeMap::new(),
             macro_defines: BTreeMap::new(),
+            new_macro_defines: HashSet::new(),
             missing_macros: HashSet::new(),
             known_replacement: HashSet::new(),
-            included: HashSet::new(),
-            parsing_macro_replacement: false,
-            parsing_macro_arg: false,
-            options,
+            disable_macro_expand: false,
+            parsing_tokens: false,
             text,
             path,
+            last_parse_error: None,
+            include: IncludeHandler::new(options),
         }
     }
 
+    pub(crate) fn set_known_macro_defines(&mut self, macro_defines: MacroDefines) {
+        assert!(self.macro_defines.is_empty());
+        self.macro_defines = macro_defines;
+    }
+
+    pub(crate) fn new_macro_defines(self) -> MacroDefines {
+        self.macro_defines
+            .iter()
+            .filter_map(|(k, v)| {
+                if self.new_macro_defines.contains(k) {
+                    Some((k.clone(), v.clone()))
+                } else {
+                    None
+                }
+            })
+            .collect()
+    }
+
+    pub fn parse_tokens<T: Parse>(&mut self, tokens: Vec<LexicalToken>) -> Result<T> {
+        let old_tokens = std::mem::replace(&mut self.tokens, tokens);
+        let old_index = self.current_token_index;
+        let old_last_parse_error = self.last_parse_error.take();
+        self.current_token_index = 0;
+
+        self.parsing_tokens = true;
+        let result = self.parse();
+        self.parsing_tokens = false;
+
+        let _ = std::mem::replace(&mut self.tokens, old_tokens);
+        self.current_token_index = old_index;
+        self.last_parse_error = old_last_parse_error;
+
+        result
+    }
+
     pub fn parse<T: Parse>(&mut self) -> Result<T> {
         let index = self.current_token_index;
         let result = T::parse(self);
-        if result.is_err() {
+        if let Err(e) = &result {
             self.current_token_index = index;
+            if self
+                .last_parse_error
+                .as_ref()
+                .map_or(true, |e0| e0.position() < e.position())
+            {
+                self.last_parse_error = Some(e.clone());
+            }
+        } else {
+            let succeeded_position = self.prev_token_end_position();
+            if self
+                .last_parse_error
+                .as_ref()
+                .map_or(false, |e| e.position() < succeeded_position)
+            {
+                self.last_parse_error = None;
+            }
         }
         result
     }
 
+    pub fn take_last_error(&mut self) -> Option<Error> {
+        self.last_parse_error.take()
+    }
+
     pub fn resume_parse<T, A>(&mut self, args: A) -> Result<T>
     where
         T: ResumeParse<A>,
@@ -109,22 +143,14 @@ impl TokenStream {
         result
     }
 
-    pub fn shared_text(&self) -> Arc<String> {
+    pub fn text(&self) -> Arc<String> {
         Arc::clone(&self.text)
     }
 
-    pub fn shared_path(&self) -> Option<Arc<PathBuf>> {
+    pub fn filepath(&self) -> Option<Arc<PathBuf>> {
         self.path.clone()
     }
 
-    pub fn filepath(&self) -> Option<PathBuf> {
-        self.tokenizer.next_position().filepath().cloned()
-    }
-
-    pub fn text(&self) -> &str {
-        self.tokenizer.text()
-    }
-
     pub fn comments(&self) -> &BTreeMap<Position, CommentToken> {
         &self.comments
     }
@@ -140,40 +166,19 @@ impl TokenStream {
         Ok(eof)
     }
 
-    // TODO: delete or rename (should only be used for EOF)
-    pub fn current_position(&self) -> Position {
-        self.tokenizer.next_position().into()
-    }
-
-    // TODO: rename (because this span may include comments and empty macros)
-    pub fn current_whitespace_token(&mut self) -> Result<WhitespaceToken> {
-        Ok(WhitespaceToken::new(
-            self.prev_token_end_position(),
-            self.next_token_start_position()?,
-        ))
-    }
-
-    pub fn enter_macro_replacement<F, T>(&mut self, f: F) -> Result<T>
+    pub fn with_macro_expand_disabled<F, T>(&mut self, f: F) -> Result<T>
     where
         F: FnOnce(&mut Self) -> Result<T>,
     {
-        self.parsing_macro_replacement = true;
+        assert!(!self.disable_macro_expand);
+
+        self.disable_macro_expand = true;
         let result = f(self);
-        self.parsing_macro_replacement = false;
+        self.disable_macro_expand = false;
         result
     }
 
-    fn enter_macro_arg<F, T>(&mut self, f: F) -> Result<T>
-    where
-        F: FnOnce(&mut Self) -> Result<T>,
-    {
-        self.parsing_macro_arg = true;
-        let result = f(self);
-        self.parsing_macro_arg = false;
-        result
-    }
-
-    fn prev_token_end_position(&mut self) -> Position {
+    pub fn prev_token_end_position(&self) -> Position {
         if let Some(i) = self.current_token_index.checked_sub(1) {
             self.tokens[i].end_position()
         } else {
@@ -181,7 +186,7 @@ impl TokenStream {
         }
     }
 
-    fn next_token_start_position(&mut self) -> Result<Position> {
+    pub fn next_token_start_position(&mut self) -> Result<Position> {
         let index = self.current_token_index;
         if index == self.tokens.len() && self.is_eof()? {
             Ok(self.tokenizer.next_position().into())
@@ -190,17 +195,14 @@ impl TokenStream {
         }
     }
 
-    fn read_token(&mut self) -> Result<Option<Token>> {
+    fn read_token(&mut self) -> Result<Option<LexicalToken>> {
         if let Some(token) = self.tokens.get(self.current_token_index).cloned() {
             self.current_token_index += 1;
 
-            if !self.parsing_macro_replacement {
-                // TODO: note comment
+            if !self.disable_macro_expand {
                 match &token {
-                    Token::Symbol(x) if x.value() == Symbol::Question => {
-                        if !self.parsing_macro_arg {
-                            return self.expand_macro_and_read_token();
-                        }
+                    LexicalToken::Symbol(x) if x.value() == Symbol::Question => {
+                        return self.expand_macro_and_read_token();
                     }
                     _ => {}
                 }
@@ -208,11 +210,19 @@ impl TokenStream {
 
             return Ok(Some(token));
         }
+        if self.parsing_tokens {
+            return Ok(None);
+        }
 
-        while let Some(token) = self.tokenizer.next().transpose()? {
+        while let Some(token) = self
+            .tokenizer
+            .next()
+            .transpose()
+            .map_err(|e| Error::tokenize_error(self, e))?
+        {
             let start_position = Position::from(token.start_position());
             let end_position = Position::from(token.end_position());
-            let token: Token = match token {
+            let token: LexicalToken = match token {
                 erl_tokenize::Token::Whitespace(_) => {
                     continue;
                 }
@@ -258,12 +268,12 @@ impl TokenStream {
             self.current_token_index += 1;
 
             match &token {
-                Token::Symbol(x) if x.value() == Symbol::Question => {
-                    if !self.parsing_macro_arg {
+                LexicalToken::Symbol(x) if x.value() == Symbol::Question => {
+                    if !self.disable_macro_expand {
                         return self.expand_macro_and_read_token();
                     }
                 }
-                Token::Symbol(x) if x.value() == Symbol::Hyphen => {
+                LexicalToken::Symbol(x) if x.value() == Symbol::Hyphen => {
                     let index = self.current_token_index;
                     self.try_handle_directives()?;
                     self.current_token_index = index;
@@ -294,7 +304,7 @@ impl TokenStream {
         (without_args, with_args)
     }
 
-    fn expand_macro_and_read_token(&mut self) -> Result<Option<Token>> {
+    fn expand_macro_and_read_token(&mut self) -> Result<Option<LexicalToken>> {
         let macro_name: MacroName = self.parse()?;
         self.expand_macro(macro_name)?;
         self.read_token()
@@ -326,8 +336,9 @@ impl TokenStream {
         let start_position = self.tokens[start_index].start_position();
         let question = QuestionSymbol::new(start_position);
 
-        let r#macro =
-            self.enter_macro_arg(|ts| Macro::parse(ts, question, macro_name.clone(), true))?;
+        let r#macro: Macro = self.with_macro_expand_disabled(|ts| {
+            ts.resume_parse((question, macro_name.clone(), true))
+        })?;
         let arity = r#macro.arity();
         assert!(arity.is_some());
 
@@ -336,7 +347,7 @@ impl TokenStream {
             let variables = define.variables.as_ref().map(|x| x.to_owned());
 
             let replacement = r#macro.expand(variables, define.replacement);
-            self.replace_tokens(start_index, replacement);
+            self.replace_tokens(start_index, &key, replacement);
             self.macros.entry(start_position).or_insert(r#macro);
         } else {
             self.expand_unknown_macro(macro_name)?;
@@ -348,15 +359,16 @@ impl TokenStream {
     fn expand_macro_without_args(
         &mut self,
         macro_name: MacroName,
-        replacement: Vec<Token>,
+        replacement: Vec<LexicalToken>,
     ) -> Result<()> {
+        let key = MacroDefineKey::new(macro_name.value().to_owned(), None);
         let start_index = self.current_token_index - 2;
         let start_position = self.tokens[start_index].start_position();
         let question = QuestionSymbol::new(start_position);
-        let r#macro = Macro::parse(self, question, macro_name, false)?;
+        let r#macro: Macro = self.resume_parse((question, macro_name, false))?;
 
         let replacement = r#macro.expand(None, replacement);
-        self.replace_tokens(start_index, replacement);
+        self.replace_tokens(start_index, &key, replacement);
         self.macros.entry(start_position).or_insert(r#macro);
 
         Ok(())
@@ -368,15 +380,14 @@ impl TokenStream {
 
         if let Some(replacement) = get_predefined_macro(macro_name.value(), start_position) {
             self.expand_macro_without_args(macro_name, replacement)
-        } else if self.parsing_macro_replacement {
+        } else if self.disable_macro_expand {
             log::debug!(
-                "Found an undefined macro {:?} in a macro replacement.",
+                "Found an undefined macro {:?} in disabling macro expansions.",
                 macro_name.value()
             );
             Ok(())
         } else {
             if !self.missing_macros.contains(macro_name.value()) {
-                // TODO: consider arity
                 log::warn!(
                     "The macro {:?} is not defined. 'EFMT_DUMMY' atom is used instead.",
                     macro_name.value()
@@ -385,22 +396,32 @@ impl TokenStream {
             }
             self.expand_macro_without_args(
                 macro_name,
-                vec![Token::from(dummy_atom(start_position))],
+                vec![LexicalToken::from(dummy_atom(start_position))],
             )
         }
     }
 
-    fn replace_tokens(&mut self, start_index: usize, mut replacement: Vec<Token>) {
+    fn replace_tokens(
+        &mut self,
+        start_index: usize,
+        key: &MacroDefineKey,
+        mut replacement: Vec<LexicalToken>,
+    ) {
         if !replacement.is_empty()
             && !self
                 .known_replacement
                 .insert((start_index, replacement.clone()))
         {
             log::warn!(
-                "A circular macro was detected. It was replaced with a dummy atom 'EFMT_DUMMY'."
+                "A circular macro {:?} was detected. It was replaced with a dummy atom 'EFMT_DUMMY'.",
+                  key.to_string(),
             );
             let start_position = self.tokens[start_index].start_position();
-            replacement = vec![Token::from(dummy_atom(start_position))];
+            replacement = vec![LexicalToken::from(dummy_atom(start_position))];
+            self.macro_defines
+                .get_mut(key)
+                .expect("unreachable")
+                .replacement = replacement.clone();
         }
 
         let unread_tokens = self.tokens.split_off(self.current_token_index);
@@ -418,6 +439,7 @@ impl TokenStream {
                 let name = x.macro_name().to_owned();
                 let define: MacroDefine = x.into();
                 let key = MacroDefineKey::new(name, define.arity());
+                self.new_macro_defines.insert(key.clone());
                 self.macro_defines.insert(key, define);
             }
             Ok(Either::B(x)) => {
@@ -428,171 +450,26 @@ impl TokenStream {
         Ok(())
     }
 
-    fn try_load_macro_defines_from_cache(
-        &self,
-        include_path: &str,
-    ) -> Option<BTreeMap<MacroDefineKey, MacroDefine>> {
-        // TODO: check file mtime
-        if let Some(cache_path) = self
-            .options
-            .include_cache_dir
-            .as_ref()
-            .map(|x| x.join(include_path))
-        {
-            // TODO: improve error handling
-            std::fs::File::open(&cache_path)
-                .map_err(|e| {
-                    log::debug!("cache file open error ({:?}): {}", cache_path, e);
-                    e
-                })
-                .ok()
-                .map(BufReader::new)
-                .and_then(|file| {
-                    serde_json::from_reader(file)
-                        .map(|entries: Vec<(String, MacroDefine)>| {
-                            entries
-                                .into_iter()
-                                .map(|(name, define)| {
-                                    (MacroDefineKey::new(name, define.arity()), define)
-                                })
-                                .collect()
-                        })
-                        .map_err(|e| {
-                            log::warn!("deserialize error ({:?}): {}", cache_path, e);
-                            e
-                        })
-                        .ok()
-                })
-        } else {
-            log::debug!("[TODO] fail1");
-            None
-        }
-    }
-
-    fn try_save_macro_defines_into_cache(
-        &self,
-        include_path: &str,
-        macro_defines: &BTreeMap<MacroDefineKey, MacroDefine>,
-    ) {
-        if let Some(cache_path) = self
-            .options
-            .include_cache_dir
-            .as_ref()
-            .map(|x| x.join(include_path))
-        {
-            // TODO: error handling
-            let _ = tempfile::NamedTempFile::new()
-                .ok()
-                .map(BufWriter::new)
-                .map(|mut file| {
-                    let entries = macro_defines
-                        .iter()
-                        .map(|(k, v)| (&k.name, v))
-                        .collect::<Vec<_>>();
-                    let saved = serde_json::to_writer(&mut file, &entries)
-                        .map_err(|e| {
-                            log::warn!("serialization error: {}", e);
-                            e
-                        })
-                        .ok()
-                        .and_then(|_| {
-                            cache_path.parent().and_then(|p| {
-                                std::fs::create_dir_all(p)
-                                    .map_err(|e| {
-                                        log::warn!("create_dir_all({:?}) error: {}", p, e);
-                                        e
-                                    })
-                                    .ok()
-                            })
-                        })
-                        .is_some();
-                    if saved {
-                        if let Err(e) = file.into_inner().expect("TODO").persist(&cache_path) {
-                            log::warn!("cannot save cache file {:?}: {}", cache_path, e);
-                        }
-                        log::debug!(
-                            "Saved a include cache for {:?} into {:?}",
-                            include_path,
-                            cache_path
-                        );
-                    }
-                });
-        }
-    }
-
     fn handle_include(&mut self, include: IncludeDirective) {
-        if self.included.contains(include.path()) {
-            return;
-        }
-        self.included.insert(include.path().to_owned());
-
-        if let Some(macro_defines) = self.try_load_macro_defines_from_cache(include.path()) {
-            log::debug!(
-                "Found {} macro definitions in {:?} (cached)",
-                macro_defines.len(),
-                include.path()
-            );
-
-            self.macro_defines.extend(macro_defines);
-            return;
-        }
-
-        let path = if let Some(path) = include.get_include_path(&self.options.include_dirs) {
-            path
-        } else {
-            log::warn!("Cannot find include file {:?}", include.path());
-            return;
-        };
-
-        let text = match std::fs::read_to_string(&path) {
-            Ok(text) => text,
-            Err(e) => {
-                log::warn!("Cannot read include file {:?}: {}", include.path(), e);
-                return;
-            }
-        };
-
-        let mut tokenizer = Tokenizer::new(text);
-        tokenizer.set_filepath(&path);
-        let mut ts = TokenStream::new(tokenizer, self.options.clone());
-        ts.macro_defines = self.macro_defines.clone();
-        match ts.parse::<Module>() {
-            Ok(_) => {
-                // TODO: Check replacement before filtering
-                let new_macro_defines = ts
-                    .macro_defines
-                    .iter()
-                    .filter(|(k, _)| !self.macro_defines.contains_key(k))
-                    .map(|(k, v)| (k.clone(), v.clone()))
-                    .collect::<BTreeMap<_, _>>();
-
-                log::debug!(
-                    "Found {} macro definitions in {:?}",
-                    new_macro_defines.len(),
-                    path
-                );
-
-                self.try_save_macro_defines_into_cache(include.path(), &new_macro_defines);
-
-                self.macro_defines.extend(ts.macro_defines);
-            }
-            Err(e) => {
-                log::warn!("Cannot parse include file {:?}: {}", include.path(), e);
-            }
-        }
+        let new_macro_defines = self
+            .include
+            .include_macro_defines(&include, &self.macro_defines);
+        self.new_macro_defines
+            .extend(new_macro_defines.keys().cloned());
+        self.macro_defines.extend(new_macro_defines);
     }
 }
 
 impl Iterator for TokenStream {
-    type Item = Result<Token>;
+    type Item = Result<LexicalToken>;
 
     fn next(&mut self) -> Option<Self::Item> {
         self.read_token().transpose()
     }
 }
 
-fn get_predefined_macro(name: &str, position: Position) -> Option<Vec<Token>> {
-    let token: Token = match name {
+fn get_predefined_macro(name: &str, position: Position) -> Option<Vec<LexicalToken>> {
+    let token: LexicalToken = match name {
         "MODULE" | "FUNCTION_NAME" => dummy_atom(position).into(),
         "LINE" | "FUNCTION_ARITY" | "OTP_RELEASE" => dummy_integer(position).into(),
         "MODULE_STRING" | "FILE" | "MACHINE" => dummy_string(position).into(),
@@ -613,22 +490,38 @@ fn dummy_string(position: Position) -> StringToken {
     StringToken::new("EFMT_DUMMY", position, position)
 }
 
-#[derive(Debug, Clone, PartialEq, Eq, PartialOrd, Ord, serde::Serialize, serde::Deserialize)]
-struct MacroDefineKey {
+#[derive(
+    Debug, Clone, PartialEq, Eq, Hash, PartialOrd, Ord, serde::Serialize, serde::Deserialize,
+)]
+pub(crate) struct MacroDefineKey {
     name: String,
     arity: Option<usize>,
 }
 
 impl MacroDefineKey {
-    fn new(name: String, arity: Option<usize>) -> Self {
+    pub(crate) fn new(name: String, arity: Option<usize>) -> Self {
         Self { name, arity }
     }
+
+    pub(crate) fn name(&self) -> &str {
+        &self.name
+    }
+}
+
+impl std::fmt::Display for MacroDefineKey {
+    fn fmt(&self, f: &mut std::fmt::Formatter) -> std::fmt::Result {
+        if let Some(arity) = self.arity {
+            write!(f, "?{}/{}", self.name, arity)
+        } else {
+            write!(f, "?{}", self.name)
+        }
+    }
 }
 
 #[derive(Debug, Clone, serde::Serialize, serde::Deserialize)]
-struct MacroDefine {
+pub(crate) struct MacroDefine {
     variables: Option<Vec<String>>,
-    replacement: Vec<Token>,
+    replacement: Vec<LexicalToken>,
 }
 
 impl From<DefineDirective> for MacroDefine {
@@ -643,7 +536,9 @@ impl From<DefineDirective> for MacroDefine {
 }
 
 impl MacroDefine {
-    fn arity(&self) -> Option<usize> {
+    pub(crate) fn arity(&self) -> Option<usize> {
         self.variables.as_ref().map(|x| x.len())
     }
 }
+
+pub(crate) type MacroDefines = BTreeMap<MacroDefineKey, MacroDefine>;
